from geocelery import Function

def xxx():
    import time
    while True:
        print("hello,world")
        time.sleep(1)

if __name__ == "__main__":


    #fbb31332-c487-4918-8652-b5555c51f6c1
    #f2237140-548d-4dbf-9e85-9b8e79bb83a5
    #5e070209-0dd7-4b4d-8e67-8ebce6a45ac5
    Function.stop("ae951335-4b20-49c2-b6ba-e0a58030251a")

    # taskid = Function.start(xxx)
    #
    # import time
    # time.sleep(5)
    #
    # print(Function.status(taskid))
    #
    # Function.stop(taskid)
    #
    # print(Function.status(taskid))



import time
from geocelery import MapReduce

def ndvi(pos):
    import dboxio
    ds = dboxio.DBoxDataset("/dboxstorage/c0/p3/GS_LT51300341992300BJC00_B30.DBOX")
    pts = ds.ReadPoints(pos)
    return pts.tolist()[0]


def g(results):
    return sum(results)

beg = time.time()

taskid = MapReduce.start([[104, 37]]*1000, ndvi, g)

rst = MapReduce.result(taskid)

print(time.time()-beg, rst)


# def task():
#     taskid = Function.start(ndvi)
#     rst = Function.result(taskid)
#     return rst
#
# beg = time.time()
#
# futures = []
# with ThreadPoolExecutor(max_workers=100) as exec:
#     for idx in range(50):
#         future_rst = exec.submit(task)
#         futures.append(future_rst)
#
# for ft in futures:
#     print(ft.result())
#
# print(time.time()-beg)








'''
用户提交的每个任务后台都有多个celery任务与之对应，token负责记录管理以上对应关系
'''
import redis
from geocelery._.settings import TOKENS_URL, BATCH_URL


class TaskToken:

    MUTEX_EXCE_PREFIX = "EXECUTE:"   #执行权key：谁先获得这个token谁执行，由获得token的任务将其删除，来阻止其他任务的执行

    CELERY_TASK_PREFIX = "TASKID:" #关联关系key：记录当其任务对应具体的celery任务的id(排队的时候是3个，执行的时候是1个)

    @classmethod
    def check_and_set(cls, taskid, celeryid):
        '''
        同一个任务对应的多个celery任务的互斥管理
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)

        exists = rdb.exists(cls.MUTEX_EXCE_PREFIX + celeryid)  #故障恢复后的任务重试
        if exists:
            return True

        #如果MUTEX_TOKEN_PREFIX+taskid存在，则返回1(获得执行权)，否则返回0(未能获得执行权)
        nums = rdb.delete(cls.MUTEX_EXCE_PREFIX + taskid)
        if nums == 0:
            return False

        #由获得执行权的任务进行初始化设置
        rdb.set(cls.MUTEX_EXCE_PREFIX + celeryid, taskid)  #把原来的token改个名字，可以防止该任务在失败重试时被拒绝
        rdb.set(cls.CELERY_TASK_PREFIX + taskid, celeryid) #记录当前负责执行taskid的celery任务是哪一个

        return True

    @classmethod
    def remove(cls, celeryid):
        '''
        任务执行结束后删除执行权key
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)
        rdb.delete(cls.MUTEX_EXCE_PREFIX + celeryid)

    @classmethod
    def setCeleryids(cls, taskid, celeryids):
        '''
        记录任务(taskid)关联的所有celery任务标识(celeryids数组)
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)
        rdb.set(cls.MUTEX_EXCE_PREFIX + taskid, None)
        rdb.set(cls.CELERY_TASK_PREFIX + taskid, ",".join(celeryids))

    @classmethod
    def getCeleryids(cls, taskid):
        '''
        获取任务(taskid)对应的所有celery任务的标识，需要操作底层任务时调用(如查询状态、获取结果)
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)
        taskids = rdb.get(cls.CELERY_TASK_PREFIX + taskid)
        if not taskids:
            return []
        return taskids.decode("utf-8").split(",")

    @classmethod
    def setMassCeleryids(cls, all_celeryids):
        '''
        批量记录用户任务与celery任务的对应关系(适合map/reduce场景)
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)
        pipe = rdb.pipeline()
        for taskid in all_celeryids:
            celeryids = all_celeryids[taskid]
            pipe.set(cls.MUTEX_EXCE_PREFIX + taskid, None)
            pipe.set(cls.CELERY_TASK_PREFIX + taskid, ",".join(celeryids))
        pipe.execute()

    @classmethod
    def removeAll(cls, taskid):
        '''
        删除redis中所有与taskid关联的信息(终止任务时调用)
        '''
        rdb = redis.StrictRedis.from_url(TOKENS_URL)
        rdb.delete(cls.MUTEX_EXCE_PREFIX + taskid)
        rdb.delete(cls.CELERY_TASK_PREFIX + taskid)


class BatchToken:

    RUNING_TAG = "r"    #正在执行celery任务标识
    COUNTS_TAG = "n"    #已完成的任务数量
    TOTALS_TAG = "t"    #批量任务中的全部任务数量
    REVOKE_TAG = "z"    #判断该批量任务是否已被客户端撤销

    @classmethod
    def reset_token(cls, taskid):
        '''
        设置revoke key，表明该任务已被终止(撤销)
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        rdb.set(taskid + ":" + cls.REVOKE_TAG, "")

    @classmethod
    def check_revoked_token(cls, taskid):
        '''
        检查有没有撤销标志key，有的话表明该任务已被终止
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        return rdb.exists(taskid + ":" + cls.REVOKE_TAG)


    @classmethod
    def set_recent_celeryid(cls, taskid, celeryid):
        '''
        记录最近一个被启动的celery任务
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        rdb.set(taskid + ":" + cls.RUNING_TAG, celeryid)

    @classmethod
    def get_recent_celeryid(cls, taskid):
        '''
        最近一个被启动的celery任务是哪一个
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        celery_id = rdb.get(taskid + ":" + cls.RUNING_TAG)
        if celery_id:
            return celery_id.decode("utf-8")
        return None


    @classmethod
    def add_finished_count(cls, taskid):
        '''
        增加任务完成数量(+1)
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        rdb.incr(taskid + ":" + cls.COUNTS_TAG)

    @classmethod
    def get_finished_count(cls, taskid):
        '''
        获取任务已完成的数量
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        _rst = rdb.get(taskid + ":" + cls.COUNTS_TAG)
        if _rst:
            return int(_rst.decode("utf-8"))
        return None

    @classmethod
    def set_totals(cls, taskid, totals):
        '''
        记录该批量任务包含的总任务量(totals)
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        rdb.set(taskid + ":" + cls.TOTALS_TAG, totals)

    @classmethod
    def get_totals(cls, taskid):
        '''
        返回该批量任务包含的总任务量
        '''
        rdb = redis.StrictRedis.from_url(BATCH_URL)
        rst = rdb.get(taskid + ":" + cls.TOTALS_TAG)
        if not rst:
            return 0
        return int(rst.decode("utf-8"))


'''
用于记录用户提交的任务中包含的所有map子任务(celery标识)/reduce子任务(celery标识)
可以将celery上所有的属于同一个任务的map、reduce子任务当作一个整体进行管理
'''
import redis
from geocelery._.settings import MAPREDUCE_URL


class MapReduceCookie:
    
    MAP_TASK_PREFIX = "MAP:"            #记录最近一个启动的map任务的celeryID
    REDUCE_TASK_PREFIX = "REDUCE:"      #记录reduce任务的celeryID
    ALL_MAP_TASKS_PREFIX = "ALLMAP:"    #记录所有map子任务的celeryID
    
    @classmethod
    def exists(cls, taskid):
        '''
        判断当前任务(taskid)是否已经开始执行
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        mp_exists = rdb.exists(cls.MAP_TASK_PREFIX + taskid)
        rd_exists = rdb.exists(cls.REDUCE_TASK_PREFIX + taskid)
        mr_exists = rdb.exists(cls.ALL_MAP_TASKS_PREFIX + taskid)
        return mp_exists or rd_exists or mr_exists

    @classmethod
    def setMapCeleryid(cls, taskid, mapid):
        '''
        记录最近一个启动的map任务(每个map任务在启动时都会调用这个函数)
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.set(cls.MAP_TASK_PREFIX + taskid, mapid)

    @classmethod
    def getMapCeleryid(cls, taskid):
        '''
        获取最新的map任务的标识
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        return rdb.get(cls.MAP_TASK_PREFIX + taskid)

    @classmethod
    def setReduceCeleryid(cls, taskid, celeryid):
        '''
        记录reduce任务的标识
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.set(cls.REDUCE_TASK_PREFIX + taskid, celeryid)

    @classmethod
    def getReduceCeleryid(cls, taskid):
        '''
        返回reduce任务的标识
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        return rdb.get(cls.REDUCE_TASK_PREFIX + taskid)

    @classmethod
    def delMapCeleryid(cls, taskid):
        '''
        从redis中删除map任务key
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.delete(cls.MAP_TASK_PREFIX + taskid)

    @classmethod
    def delReduceCeleryid(cls, taskid):
        '''
        从redis中删除reduce任务key
        :param taskid:
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.delete(cls.REDUCE_TASK_PREFIX + taskid)

    @classmethod
    def setMapTaskids(cls, taskid, mapids):
        '''
        记录所有属于taskid的map任务标识
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.set(cls.ALL_MAP_TASKS_PREFIX + taskid, ",".join(mapids))

    @classmethod
    def getMapTaskids(cls, taskid):
        '''
        返回所有的map任务标识
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        taskids = rdb.get(cls.ALL_MAP_TASKS_PREFIX + taskid)
        if not taskids:
            return []

        return taskids.decode("utf-8").split(",")

    @classmethod
    def remove(cls, taskid):
        '''
        从redis中删除所有与taskid对应的信息
        '''
        rdb = redis.StrictRedis.from_url(MAPREDUCE_URL)
        rdb.delete(cls.MAP_TASK_PREFIX + taskid)
        rdb.delete(cls.REDUCE_TASK_PREFIX + taskid)
        rdb.delete(cls.ALL_MAP_TASKS_PREFIX + taskid)


import uuid
import time
import inspect
from concurrent.futures import ThreadPoolExecutor
from celery import chord, group
from celery.result import AsyncResult
from celery.states import PENDING, RECEIVED, STARTED, SUCCESS, REVOKED
from geocelery._.application import task_priority, celery_batch, TaskPriority
from geocelery._.application import celery_app, celery_map, celery_reduce, TaskType, celery_function
from geocelery._.tokens import TaskToken, BatchToken
from geocelery._.cookies import MapReduceCookie


class TaskNotFound(Exception):
    '''
    未找到对应的任务
    '''
    pass


class BadFunctionType(Exception):
    '''
    不支持的函数类型
    '''
    pass


class TaskNotSucceed(Exception):
    '''
    任务没有成功
    '''
    pass


def trim_indent(func_text):
    '''
    删除函数定义文本前的缩进
    '''
    func_lines = func_text.split("\n")
    indents = func_lines[0].index("def")
    return "\n".join([func_line[indents:] for func_line in func_lines])


def is_bounded_func(func):
    '''
    判断被调用函数是否是类方法
    '''
    return hasattr(func, '__cls__')



class State:
    '''
    任务状态
    '''
    PENDING = 'PENDING'
    RECEIVED = 'RECEIVED'
    STARTED = 'STARTED'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'
    REVOKED = 'REVOKED'
    REJECTED = 'REJECTED'
    RETRY = 'RETRY'
    IGNORED = 'IGNORED'


class Function:
    '''
    普通函数执行类
    '''
    
    @classmethod
    def start(cls, target, args=None, kwargs=None, tasktype=TaskType.B):
        '''
        开始在celery上执行target对应的任务
        :param target: 需要被提交导平台上执行的函数
        :param args: 传到target函数的位置参数
        :param kwargs: 传到target函数的关键字参数
        :param tasktype: 任务类型，默认为普通类型(TaskType.B)
        :return: 当前任务的taskid
        '''
        if not callable(target) or is_bounded_func(target):
            raise BadFunctionType("target:%s must be unbounded function!" % str(target))

        func_name = target.__name__
        func_text = trim_indent(inspect.getsource(target))

        taskid = str(uuid.uuid4())

        hd_id, md_id, lw_id = taskid + ":a", taskid + ":b", taskid + ":c"
        TaskToken.setCeleryids(taskid, [hd_id, md_id, lw_id])

        priorities = task_priority(tasktype)

        celery_function.apply_async(
            task_id=hd_id,
            exchange="celery",
            routing_key=TaskType.A,
            priority=priorities[0],
            args=[taskid, func_name, func_text, args, kwargs]
        )
        celery_function.apply_async(
            task_id=md_id,
            exchange="celery",
            routing_key=TaskType.B,
            priority=priorities[1],
            args=[taskid, func_name, func_text, args, kwargs]
        )
        celery_function.apply_async(
            task_id=lw_id,
            exchange="celery",
            routing_key=TaskType.C,
            priority=priorities[2],
            args=[taskid, func_name, func_text, args, kwargs]
        )

        return taskid

    @classmethod
    def stop(cls, taskid):
        '''
        终止当前任务
        成功返回True，否则返回False
        '''
        taskids = TaskToken.getCeleryids(taskid)
        for taskid in taskids:
            celery_app.control.revoke(taskid, terminate=True)
        TaskToken.removeAll(taskid)
        return True

    @classmethod
    def status(cls, taskid):
        '''
        返回任务状态
        '''
        taskids = TaskToken.getCeleryids(taskid)
        if not taskids:
            return None
        taskid = taskids[0]
        async_rst = AsyncResult(taskid)
        return async_rst.status

    @classmethod
    def result(cls, taskid, timeout=None):
        '''
        返回任务结果
        '''
        timeout = timeout or 3600*24*7

        while timeout > 0:
            taskids = TaskToken.getCeleryids(taskid)
            if not taskids:
                return None

            if len(taskids) > 1:
                time.sleep(1)
                timeout = timeout - 1
                continue

            celery_rst = AsyncResult(taskids[0], app=celery_app)
            asyrst = celery_rst.get(timeout=timeout)
            return asyrst.getResult()

        raise TimeoutError()



class BatchFunction:
    '''
    批量任务提交接口
    '''
    @classmethod
    def start(cls, target, dataset, tasktype=TaskType.B):
        '''

        :param target:
        :param dataset:
        :param tasktype:
        :return:
        '''
        if not callable(target) or is_bounded_func(target):
            raise BadFunctionType("target:%s must be unbounded function!" % str(target))

        func_name = target.__name__
        func_text = trim_indent(inspect.getsource(target))

        taskid = str(uuid.uuid4())

        celery_batch.apply_async(
            exchange="celery",
            routing_key=tasktype,
            priority=TaskPriority.HIGH,
            args=[taskid, tasktype, func_name, func_text, dataset]
        )

        BatchToken.set_totals(taskid, len(dataset))

        return taskid


    @classmethod
    def status(cls, taskid):
        '''
        批量任务的状态：如果是STARTED，会同时返回进度
        '''
        if BatchToken.check_revoked_token(taskid):
            return REVOKED

        celeryid = BatchToken.get_recent_celeryid(taskid)
        if not celeryid:
            return PENDING

        celery_status = AsyncResult(celeryid).status
        if celery_status != STARTED and celery_status != SUCCESS:
            return celery_status

        nums = BatchToken.get_finished_count(taskid) or 0
        totals = BatchToken.get_totals(taskid)
        if nums < totals:
            return STARTED + ":" + str(nums) + "/" + str(totals)

        return SUCCESS


    @classmethod
    def stop(cls, taskid):
        '''
        终止taskdi对应的任务
        '''
        BatchToken.reset_token(taskid)


class MapReduce:
    '''
    map/reduce任务类型
    '''

    @classmethod
    def start(cls, dataset, map_func, reduce_func, tasktype=TaskType.B, ignore_error=False):
        '''
        启动map/reduce任务
        :param dataset: 待处理的数据集(数组/Array)
        :param map_func: map函数，参数是dataset中的一个data
        :param reduce_func: reduce函数，参数是所有map_func返回的结果的集合(数组/Array)
        :param tasktype: 任务类型，默认为B类型
        :param ignore_error: 是否忽略map错误
        :return: 返回任务标识taskid
        '''
        if not callable(map_func) or is_bounded_func(map_func):
            raise BadFunctionType("map_func:%s must be unbounded function!" % map_func.__name__)

        if not callable(reduce_func) or is_bounded_func(reduce_func):
            raise BadFunctionType("reduce_func:%s must be unbounded function!" % reduce_func.__name__)

        map_func_name = map_func.__name__
        map_func_text = trim_indent(inspect.getsource(map_func))

        reduce_func_name = reduce_func.__name__
        reduce_func_text = trim_indent(inspect.getsource(reduce_func))

        priorities = task_priority(tasktype)

        taskid = str(uuid.uuid4())

        map_tasks = []
        map_taskids = []
        map_celeryids = {}


        def _map_task(taskid, idx, ignore_error, data, map_func_name, map_func_text):
            map_task_id = taskid + ":m-" + str(idx)
            map_hd_id, map_md_id, map_lw_id = "%s-a" % map_task_id, "%s-b" % map_task_id, "%s-c" % map_task_id

            map_celeryids[map_task_id] = [map_hd_id, map_md_id, map_lw_id]

            map_taskids.extend([map_hd_id, map_md_id, map_lw_id])

            h_map_task = celery_map.signature(
                (taskid, map_task_id, ignore_error, data, map_func_name, map_func_text),
                task_id=map_hd_id,
                exchange="celery",
                routing_key=TaskType.A,
                priority=priorities[0]
            )

            m_map_task = celery_map.signature(
                (taskid, map_task_id, ignore_error, data, map_func_name, map_func_text),
                task_id=map_md_id,
                exchange="celery",
                routing_key=TaskType.B,
                priority=priorities[1]
            )

            l_map_task = celery_map.signature(
                (taskid, map_task_id, ignore_error, data, map_func_name, map_func_text),
                task_id=map_lw_id,
                exchange="celery",
                routing_key=TaskType.C,
                priority=priorities[2]
            )

            return h_map_task, m_map_task, l_map_task

        all_futures = []
        with ThreadPoolExecutor(max_workers=100) as executor:
            for idx, data in enumerate(dataset):
                future = executor.submit(_map_task, taskid, idx, ignore_error, data, map_func_name, map_func_text)
                all_futures.append(future)

        for ft in all_futures:
            sub_tasks = ft.result()
            map_tasks.extend(sub_tasks)

        TaskToken.setMassCeleryids(map_celeryids)

        MapReduceCookie.setMapTaskids(taskid, map_taskids)

        reduce_hd_id, reduce_md_id, reduce_lw_id = "%s:r-a" % taskid, "%s:r-b" % taskid, "%s:r-c" % taskid
        TaskToken.setCeleryids(taskid, [reduce_hd_id, reduce_md_id, reduce_lw_id])

        reduce_a = celery_reduce.signature(
            (taskid, reduce_func_name, reduce_func_text),
            task_id=reduce_hd_id,
            exchange="celery",
            routing_key=TaskType.A,
            priority=priorities[0] + 1
        )

        reduce_b = celery_reduce.signature(
            (taskid, reduce_func_name, reduce_func_text),
            task_id=reduce_md_id,
            exchange="celery",
            routing_key=TaskType.B,
            priority=priorities[1] + 1
        )

        reduce_c = celery_reduce.signature(
            (taskid, reduce_func_name, reduce_func_text),
            task_id=reduce_lw_id,
            exchange="celery",
            routing_key=TaskType.C,
            priority=priorities[2] + 1
        )

        r_task = group(reduce_a, reduce_b, reduce_c)

        mr_task = chord(map_tasks, r_task)

        mr_task.apply_async()

        return taskid


    @classmethod
    def status(cls, taskid):
        '''
        任务状态
        '''
        map_celery_id = MapReduceCookie.getMapCeleryid(taskid)
        if map_celery_id:
            map_task = AsyncResult(map_celery_id)
            map_status = map_task.status
            if map_status in [PENDING, RECEIVED, STARTED]:
                return map_status

        red_celery_id = MapReduceCookie.getReduceCeleryid(taskid)
        if not red_celery_id:
            if MapReduceCookie.exists(taskid):
                return PENDING
            return None

        reduce_task = AsyncResult(red_celery_id)
        reduce_status = reduce_task.status
        if reduce_status in [PENDING, RECEIVED]:
            return STARTED
        return reduce_status


    @classmethod
    def result(cls, taskid, timeout=None):
        '''
        返回任务结果
        '''
        timeout = timeout or 3600*24*7

        while timeout > 0:
            reduce_celery_id = MapReduceCookie.getReduceCeleryid(taskid)
            if not reduce_celery_id:
                time.sleep(1)
                timeout = timeout - 1
                if not MapReduceCookie.exists(taskid):
                    raise TaskNotFound()
                continue

            reduce_rst = AsyncResult(reduce_celery_id)
            asyrst = reduce_rst.get(timeout=timeout)
            return asyrst.getResult()

        raise TimeoutError()


    @classmethod
    def stop(cls, taskid):
        '''
        终止任务
        '''
        reduce_taskids = TaskToken.getCeleryids(taskid)
        for taskid in reduce_taskids:
            celery_app.control.revoke(taskid, terminate=True)

        map_taskids = MapReduceCookie.getMapTaskids(taskid)
        for taskid in map_taskids:
            celery_app.control.revoke(taskid, terminate=True)

        MapReduceCookie.remove(taskid)




#消息中间件的配置信息
# BROKER_URL = [
#     'pyamqp://guest:guest@10.0.86.24:5672//',
#     'pyamqp://guest:guest@10.0.86.92:5672//',
#     'pyamqp://guest:guest@10.0.86.121:5672//'
# ]

BROKER_URL = 'pyamqp://guest:guest@10.0.85.230:5672//'

REDIS_HOST = "10.0.85.231"

#暂存任务的执行结果
RESULT_URL = "redis://%s:6379/1"%REDIS_HOST

#存储管理单个任务对应的celery任务标识
TOKENS_URL = "redis://%s:6379/2"%REDIS_HOST

#存储管理MAP/Reduce任务对应的celery任务标识
MAPREDUCE_URL = "redis://%s:6379/3"%REDIS_HOST

#负责记录批量任务的标识
BATCH_URL = "redis://%s:6379/4"%REDIS_HOST

#任务失败之后的重试等待事件
DEFAULT_RETRY_DELAY = 0

#每个失败的任务最多重试次数
DEFAULT_RETRY_TIMES = 3




from dboxapi import *

c = DBoxQuery()

pt = {"type": "point",
      "coordinates": [
          112.03653,
          38.897676
      ]
      }

q1 = LogicalFilter(
    LeafTerms("date_year", [2002, 2003]),
    LeafTerm("level", 0),
    LeafTerms("date_month", [5, 6, 7, 8])
)

q = BooleanQuery(q1)

print("检索条件为：", str(q))

jdata, code = c.m_query(q)

if jdata is not None:
    print("\n符合条件记录数：", jdata["count"])

    _id = jdata["items"][0]["id"]

    jitem, code = c.m_info(_id)
    if jitem is not None:
        print("====")
        test_jdata = jitem


from geocelery import MapReduce
import dboxmr


def ndvi_map(args):
    import numpy as np
    import dboxio

    x, y, kwargs = args

    band3_path = kwargs["band3"]
    band4_path = kwargs["band4"]
    ndvi_ds_path = kwargs["dstfile"]
    dst_nodata = kwargs["nodata"]

    nir_dsh = dboxio.DBoxDataset(band3_path)
    nir_ds = nir_dsh.GetGridAsGDALDataset(x, y)
    if not nir_ds:
        return None

    cols, rows = nir_dsh.GetGridSize(x, y)

    nir_band = nir_ds.GetRasterBand(1)
    nir_nodata = nir_band.GetNoDataValue()

    np_nir = nir_band.ReadAsArray(0, 0, cols, rows)
    nir_mask = np.equal(np_nir, nir_nodata)

    red_dsh = dboxio.DBoxDataset(band4_path)
    red_ds = red_dsh.GetGridAsGDALDataset(x, y)
    if not red_ds:
        return None

    red_band = red_ds.GetRasterBand(1)
    red_nodata = red_band.GetNoDataValue()

    np_red = red_band.ReadAsArray(0, 0, cols, rows)
    red_mask = np.equal(np_red, red_nodata)

    nodata_mask = np.logical_or(nir_mask, red_mask)

    np_nir_as32 = np_nir.astype(np.float32)
    np_red_as32 = np_red.astype(np.float32)

    result = (np_nir_as32 - np_red_as32) / (np_nir_as32 + np_red_as32)

    result = np.where(nodata_mask, dst_nodata, result)
    result = result.astype(np.float32)

    #     gdal.SetConfigOption( "DBoxDisableCache", "1" )
    ndvi_ds = dboxio.DBoxDataset(ndvi_ds_path, dboxio.GA_Update)
    ndvi_dsh = ndvi_ds.GetGridAsGDALDataset(x, y)
    if not ndvi_dsh:
        return None

    ndvi_band = ndvi_dsh.GetRasterBand(1)
    ndvi_band.WriteArray(result, 0, 0)
    ndvi_dsh.FlushCache()
    ndvi_ds.Close()
    del ndvi_ds

    ndvi_ds = dboxio.DBoxDataset(ndvi_ds_path, dboxio.GA_ReadOnly)
    ndvi_dsh = ndvi_ds.GetGridAsGDALDataset(x, y)
    if not ndvi_dsh:
        return None
    ndvi_band = ndvi_dsh.GetRasterBand(1)
    print(ndvi_band.GetStatistics(0, 1))

    return ndvi_ds.GetGridFile(x, y)


def ndvi_reduce(all_tiffs):
    return [tiff for tiff in all_tiffs if tiff is not None]


def calc_lst_ndvi(band3, band4, **kwargs):
    '''
    利用 Landsat 影像计算 ndvi，需要 urllib3
    '''
    if not band3 or not band4:
        return None, None

    import gdalconst
    from dboxapi import DBoxCCData, DBoxQuery

    band3 = DBoxCCData(band3).fullname()
    band4 = DBoxCCData(band4).fullname()

    cquery = DBoxQuery()
    all_tiles = []

    dst_nodata = -10
    jdata, status = cquery.ds_tmp(band3, gdalconst.GDT_Float32, 1, dst_nodata)
    # 调用远程服务创建一个零时文件对象
    #     print(jdata)
    if status != 200:
        print("-----------------")
        return None, None

    xoff = jdata["xtiles"]
    yoff = jdata["ytiles"]

    ndvi_ds_path = DBoxCCData(jdata["objectid"]).fullname()

    for idx in range(len(xoff)):
        for idy in range(len(yoff)):
            all_tiles.append([idx, idy, {
                "band3": band3,
                "band4": band4,
                "dstfile": ndvi_ds_path,
                "nodata": dst_nodata
            }])

    taskid = MapReduce.start(all_tiles, ndvi_map, ndvi_reduce)
    rst = MapReduce.result(taskid)

    return jdata["objectid"], [DBoxCCData(tiff).objectid() for tiff in rst]


if __name__ == "__main__":
    band3 = "c0/p3/GS_LT51300341992300BJC00_B30.DBOX"
    band4 = "c0/p4/GS_LT51300341992300BJC00_B40.DBOX"

    rst = calc_lst_ndvi(band3, band4)

    print(rst)



from geocelery import Function

def xxx():
    import os
    if os.path.exists("/mnt/celery/retry_ces.txt"):
        return "hello,world"
    else:
        open("/mnt/celery/retry_ces.txt", "w").close()
        raise Exception()

if __name__ == "__main__":
    taskid = Function.start(xxx)

    print(Function.status(taskid)) #重试之后成功，状态为RETRY
    print(Function.result(taskid))








from geocelery import MapReduce

def f(num):
    if num%3==0:
        raise Exception("xxx")
    return num

def g(nums):
    return sum(nums)


if __name__ == "__main__":
    taskid = MapReduce.start(range(1, 11), f, g, ignore_error=True)
    rst = MapReduce.result(taskid)
    print(rst, rst==sum([x for x in range(1,11) if x%3]))





from geocelery import MapReduce

def f(num):
    return num * num

def g(nums):
    return sum(nums)

if __name__ == "__main__":
    taskid = MapReduce.start(range(1, 11), f, g)
    rst = MapReduce.result(taskid)
    print(rst, rst==sum([x*x for x in range(1,11)]))


from geocelery import Function

def f1():
    time.sleep(1)

def f2():
    return 1/0


if __name__ == "__main__":
    taskid = Function.start(f1)
    print(taskid)

    rst = Function.result(taskid)
    print(rst)


from geocelery import Function, State

def add(x, y):
    import time
    time.sleep(3)
    return x + y


if __name__ == "__main__":
    taskid = Function.start(add, args=(1, 1)) #add(1, 1)

    print(Function.status(taskid))

    import time
    time.sleep(5)

    print(Function.status(taskid) == State.SUCCESS)

    print(Function.result(taskid))



import time
from celery.states import SUCCESS
from geocelery import BatchFunction

def xxx(x):
    import time
    print("------------%d------------"%x)
    time.sleep(10)
    print("++++++++++++%d++++++++++++"%x)

taskid = BatchFunction.start(xxx, range(100))

beg = time.time()
while True:
    status = BatchFunction.status(taskid)
    if status == SUCCESS:
        print(time.time()-beg)
        break
    print(status)
    time.sleep(1)


from setuptools import setup, find_packages

setup(
    name="geocelery",
    version="1.0",
    author="zhouxiaohua",
    author_email="zhouxiaohua@cnic.cn",
    packages=find_packages(),
    install_requires=["celery", "redis"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



from geocelery._.tasks import Function, MapReduce, State, BatchFunction, TaskNotFound
from geocelery._.application import TaskType

__all__ = [
    "Function", "MapReduce", "State", "TaskType", "BatchFunction", "DuplicateResult", "RetryResult", "FinalResult", "TaskNotFound"
]

class FinalResult:
    def __init__(self, taskid, celery_rst):
        self._id = taskid
        self._rst = celery_rst

    def getResult(self):
        return self._rst

    def getID(self):
        return self._id

    def __repr__(self):
        return str(self._rst)

class DuplicateResult:
    '''
    重复任务的结果
    '''
    def __init__(self, taskid):
        self.taskid = taskid

    def __repr__(self):
        return "DUPLICATE"

    def __str__(self):
        return "DUPLICATE"

    def __unicode__(self):
        return u"DUPLICATE"


class RetryResult:
    '''
    重试之后依然失败的结果
    '''
    def __init__(self, taskid):
        self.taskid = taskid

    def __repr__(self):
        return "RETRY"

    def __str__(self):
        return "RETRY"

    def __unicode__(self):
        return u"RETRY"



from geocelery import Function, TaskType

def xxx(idx):
    print("**********AAA_%d**********"%idx)
    return idx*2

def yyy(idx):
    import time
    time.sleep(10)
    # open("/celery_data/%d_BBB_%d" % (idx, int(time.time())), "w").close()
    return "BBB_%d"%idx

def zzz(idx):
    import time
    time.sleep(10)
    # open("/celery_data/%d_CCC_%d" % (idx, int(time.time())), "w").close()
    return "CCC_%d"%idx

def submit_b():
    for idx in range(1000):
        Function.start(yyy, args=(idx,), tasktype=TaskType.B)

def submit_c():
    for idx in range(1000):
        Function.start(zzz, args=(idx,), tasktype=TaskType.C)

if __name__ == "__main__":
    import time, threading
    threading.Thread(target=submit_b).start()
    threading.Thread(target=submit_c).start()

    time.sleep(30)  #等待10秒，让B/C类开始执行，然后提交A类任务

    for idx in range(10):
        beg = time.time()
        taskid = Function.start(xxx, args=(idx,), tasktype=TaskType.A)
        rst = Function.result(taskid)
        print(time.time()-beg, rst)
        time.sleep(3)






from celery import Celery, platforms
from geocelery._.settings import BROKER_URL, RESULT_URL


class TaskType:
    '''
    任务类型： A：高优先级任务； B：普通任务； C：低优先级任务
    '''
    A, B, C = "HIGH_PRIORITY", "MID_PRIORITY", "LOW_PRIORITY"

class TaskPriority:
    '''
    任务优先级：最低0，最高255
    '''
    HIGH, MID, LOW = 30, 20, 10


def task_priority(task_type):
    if task_type == TaskType.A:
        return [TaskPriority.HIGH, TaskPriority.MID, TaskPriority.HIGH]

    if task_type == TaskType.B:
        return [TaskPriority.MID, TaskPriority.HIGH, TaskPriority.LOW]

    if task_type == TaskType.C:
        return [TaskPriority.LOW, TaskPriority.LOW, TaskPriority.MID]


celery_app = Celery('tasks')

platforms.C_FORCE_ROOT = True   #for root user

celery_app.conf.update(
    BROKER_URL=BROKER_URL,
    BROKER_POOL_LIMIT=None,
    CELERYD_SEND_EVENTS=True,
    CELERY_TRACK_STARTED=True,
    CELERYD_PREFETCH_MULTIPLIER=1,
    CELERY_TASK_SERIALIZER='pickle',
    CELERY_RESULT_BACKEND=RESULT_URL,
    CELERY_RESULT_SERIALIZER='pickle',
    CELERY_ACCEPT_CONTENT=['pickle']
)

@celery_app.task(name='tasks.add')
def add(x, y):
    return x + y

@celery_app.task(name='tasks.celery_batch')
def celery_batch(taskid, tasktype, func_name, func_text, dataset):
    pass

@celery_app.task(bind=True, name='tasks.exec_function')
def celery_function(self, taskid, func_name, func_text, *args, **kwargs):
    pass


@celery_app.task(bind=True, name='tasks.celery_map')
def celery_map(self, taskid, m_taskid, ignore_error, data, func_name, func_text):
    pass

@celery_app.task(bind=True, name='tasks.celery_reduce')
def celery_reduce(self, results, taskid, func_name, func_text):
    pass
        
